# -*- encoding: utf-8 -*-
from __future__ import unicode_literals

import logging
from datetime import datetime

from sandbox import sdk2
from sandbox.common.types import resource as ctr
from sandbox.projects.common import binary_task
from sandbox.projects.avia.base import AviaBaseTask
from sandbox.projects.avia_resources.resources import AVIA_WIZARD_DIRECTION_NATIONAL
from sandbox.projects.Travel.resources import dicts

log = logging.getLogger(__name__)


class AviaStatisticsUpdateRouteCrosslinks(binary_task.LastBinaryTaskRelease, AviaBaseTask):
    class Requirements(sdk2.Requirements):
        cores = 1  # exactly 1 core
        disk_space = 128  # 128 Megs or less
        ram = 128  # 128 Megs or less

        class Caches(sdk2.Requirements.Caches):
            pass  # means that task do not use any shared caches

    class Parameters(sdk2.Parameters):
        SPECIFIC_DAY_KEY = 'specific_day'
        TODAY_KEY = 'today'

        # binary task release parameters
        ext_params = binary_task.binary_release_parameters(stable=True)

        with sdk2.parameters.Group('Yt parameters'):
            yt_proxy = sdk2.parameters.String('Proxy', default='hahn', required=True)
            yt_token = sdk2.parameters.YavSecret('YT token', default='sec-01dfxmszhq27tk66hm107d0ycd', required=True)

        with sdk2.parameters.Group('YDB parameters'):
            ydb_cluster = sdk2.parameters.String('Cluster', default='ydb-ru-prestable.yandex.net:2135', required=True)
            ydb_database = sdk2.parameters.String('Database', default='/ru-prestable/ticket/testing/avia_statistics', required=True)
            ydb_table = sdk2.parameters.String('Table', default='route_crosslinks', required=True)
            ydb_token = sdk2.parameters.YavSecret('YDB token', default='sec-01e149a59chjhzr2c1f6wcqdef', required=True)
            ydb_batch_size = sdk2.parameters.Integer(
                'YDB batch size',
                required=True,
                default=1000,
            )

        with sdk2.parameters.Group('price-index parameters'):
            price_index_host = sdk2.parameters.String(
                'price-index host',
                required=True,
                default='http://price-index.production.avia.yandex.net',
            )
            price_index_timeout = sdk2.parameters.Integer(
                'price-index request timeout (seconds)',
                required=True,
                default=5,
            )
            window_size = sdk2.parameters.Integer(
                'Price searching window size in the future',
                required=True,
                default=30,
            )
            price_index_batch_size = sdk2.parameters.Integer(
                'price-index batch size',
                required=True,
                default=100,
            )
            price_index_throttling_ms = sdk2.parameters.Float(
                'price-index throttling between requests (milliseconds)',
                required=True,
                default=1000,
            )

        landing_routes_table = sdk2.parameters.String('YT table with landing routes', default='//home/avia/avia-statistics/landing-routes', required=True)
        landing_route_weights_table = sdk2.parameters.String('YT table with landing routes', default='//home/avia/avia-statistics/landing-route-weights', required=True)

        with sdk2.parameters.RadioGroup('Crosslinks') as crosslinks:
            crosslinks.values['generate_new'] = crosslinks.Value('Generate new crosslinks', default=True)
            crosslinks.values['yt_cache'] = crosslinks.Value('Read from YT')

            with crosslinks.value['yt_cache']:
                crosslinks_table = sdk2.parameters.String('YT crosslinks table', default='//home/avia/avia-statistics/landing-route-crosslinks', required=True)

        with sdk2.parameters.RadioGroup('Run type') as log_type:
            log_type.values[TODAY_KEY] = log_type.Value('Run for today', default=True)
            log_type.values[SPECIFIC_DAY_KEY] = log_type.Value('Run for a specific day')

            with log_type.value[SPECIFIC_DAY_KEY]:
                date = sdk2.parameters.StrictString(
                    'Date YYYY-MM-DD',
                    regexp="\d{4}-\d{2}-\d{2}",
                    required=True,
                )

    def on_execute(self):
        super(AviaStatisticsUpdateRouteCrosslinks, self).on_execute()
        import ydb

        from travel.avia.library.python.lib_yt.client import configured_client
        from travel.avia.library.python.ydb.session_manager import YdbSessionManager
        from travel.library.python.dicts.avia.direction_national_repository import DirectionNationalRepository
        from travel.library.python.dicts.avia.settlement_repository import SettlementRepository
        from travel.library.python.dicts.avia.near_direction_repository import NearDirectionRepository
        from travel.avia.avia_statistics.route_crosslinks_updater.lib.updater import RouteCrosslinksUpdater
        from travel.avia.avia_statistics.route_crosslinks_updater.lib.table import RouteCrosslinksTable
        from travel.avia.avia_statistics.lib.settlements_geo_index import SettlementsGeoIndex
        from travel.avia.avia_statistics.landing_routes import get_landing_routes, get_route_weights
        from travel.avia.avia_statistics.services.price_index.client import Client as PriceIndexClient

        if self.Parameters.log_type == self.Parameters.TODAY_KEY:
            target_date = datetime.now().date()
        else:
            target_date = datetime.strptime(str(self.Parameters.date), '%Y-%m-%d').date()

        settlement_repository = SettlementRepository()
        settlement_resource_path = self._get_latest_resource_version_path(dicts.TRAVEL_DICT_AVIA_SETTLEMENTS_PROD)
        settlement_repository.load_from_file(settlement_resource_path)

        direction_national_repository = DirectionNationalRepository()
        direction_national_resource_path = self._get_latest_resource_version_path(AVIA_WIZARD_DIRECTION_NATIONAL)
        direction_national_repository.load_from_file(direction_national_resource_path)

        near_direction_repository = NearDirectionRepository()
        near_direction_resource_path = self._get_latest_resource_version_path(dicts.TRAVEL_DICT_AVIA_NEAR_DIRECTION_PROD)
        near_direction_repository.load_from_file(near_direction_resource_path)

        settlements_geo_index = SettlementsGeoIndex(direction_national_repository, settlement_repository)

        yt_client = configured_client(self.Parameters.yt_proxy, self.Parameters.yt_token.data()['token'])

        ydb_config = ydb.DriverConfig(
            endpoint=self.Parameters.ydb_cluster,
            database=self.Parameters.ydb_database,
            auth_token=self.Parameters.ydb_token.data()['token']
        )

        session_manager = YdbSessionManager(ydb_config)
        route_crosslinks_table = RouteCrosslinksTable(
            session_manager,
            self.Parameters.ydb_database,
            self.Parameters.ydb_table,
        )
        landing_routes = list(get_landing_routes(yt_client, self.Parameters.landing_routes_table))[::-1]
        routes_weights = get_route_weights(yt_client, self.Parameters.landing_route_weights_table)
        price_index_client = PriceIndexClient(self.Parameters.price_index_host)
        route_crosslinks_updater = RouteCrosslinksUpdater(
            yt_client,
            landing_routes,
            routes_weights,
            near_direction_repository,
            settlements_geo_index,
            route_crosslinks_table,
            price_index_client,
            price_index_batch_size=self.Parameters.price_index_batch_size,
            price_index_timeout=self.Parameters.price_index_timeout,
            ydb_batch_size=self.Parameters.ydb_batch_size,
            throttle_time_ms=self.Parameters.price_index_throttling_ms,
        )
        if self.Parameters.crosslinks == 'generate_new':
            route_crosslinks_updater.update(target_date, window_size=self.Parameters.window_size, yt_crosslinks_table=None)
        else:
            route_crosslinks_updater.update(target_date, window_size=self.Parameters.window_size, yt_crosslinks_table=self.Parameters.crosslinks_table)

    def _get_latest_resource_version_path(self, resource_class):
        resource = resource_class.find(state=ctr.State.READY).first()
        resource_data = sdk2.ResourceData(resource)
        return resource_data.path.as_posix()
