# coding=utf-8

import os
import re

from sandbox.projects.market.infra import (
    UnstripBinary
)
from sandbox.sdk2 import (
    Task,
    Vault,
    parameters,
    ResourceData,
    svn
)
from sandbox.sdk2.helpers import (
    subprocess as sp,
    ProcessLog
)
from sandbox.sdk2.ssh import Key
from sandbox.sdk2.path import Path
from sandbox.projects.market.infra.helpers.gpg_key import GpgKey
import sandbox.projects.common.gnupg
import sandbox.projects.common.debpkg


class MarketBuildDebPackageTestTask(Task):
    """
    Build bebian package with debuild
    """

    _KEY_OWNER = 'MARKET'

    class Parameters(Task.Parameters):
        module_path = parameters.String(
            'Module path',
            description='Example: market/backctld',
            required=True
        )
        branch_path = parameters.String(
            'Branch path',
            description='Example: market/idx/publisher/2018.1.1. If empty then trunk',
            default_value=None
        )
        revision = parameters.String(
            'Revision',
            description='If empty then HEAD',
            default_value=None
        )
        run_tests = parameters.Bool(
            'Run tests',
            description='If not checked then add DEB_BUILD_OPTIONS=nocheck option for debuild',
            default_value=True,
            required=True
        )
        package_version = parameters.String(
            'Version of package',
            required=True
        )
        dist_repos = parameters.String(
            'Repository',
            description='Repository for upload command',
            default_value=None,
            required=False
        )
        deb_user = parameters.String(
            'Debian user',
            required=True,
            description='User that will sign deb package'
        )
        deb_mail = parameters.String(
            'Debian email',
            required=True,
            description='Email of debian user'
        )
        ssh_key_name = parameters.String(
            'SSH key name in Sandbox Vault',
            group='GPG',
            required=True
        )
        gpg_private_key_name = parameters.String(
            'Private GPG key in Sandbox Vault',
            group='GPG',
            required=True
        )
        gpg_public_key_name = parameters.String(
            'Public GPG key in Sandbox Vault',
            group='GPG',
            required=True
        )
        yt_token_key_name = parameters.String(
            'YT token in Sandbox Vault',
            group='YT',
            required=False
        )
        build_with_dist = parameters.Bool(
            'Do ya make with dist option',
            default_value=False,
            group='Build'
        )
        generate_unstrip_binary = parameters.Bool(
            'Generate unstrip binary',
            default_value=False,
            group='C++ build settings'
        )

    def __make_debian_changelog(self, version, module_path):
        with ProcessLog(self, logger="dch_log") as pl:
            environment = os.environ.copy()
            environment['DEBEMAIL'] = '{user} <{mail}>'.format(
                user=self.Parameters.deb_user,
                mail=self.Parameters.deb_mail
            )
            sp.check_call(
                u'dch -b -v "{v}" --distribution unstable new release {v}'.format(
                    v=version
                ),
                shell=True,
                env=environment,
                cwd=module_path,
                stdout=pl.stdout,
                stderr=sp.STDOUT
            )

    def __debuild(self, module_path, run_test=True, out_dir=None, yt_token_key_name=None, build_with_dist=False):
        with ProcessLog(self, logger="debuild_log") as pl:
            environment = os.environ.copy()
            ya_make_opts = []
            if not run_test:
                environment['DEB_BUILD_OPTIONS'] = 'nocheck'
            if out_dir:
                ya_make_opts.append(
                    '-o {out_dir}'.format(out_dir=out_dir)
                )
            if yt_token_key_name:
                environment['YT_TOKEN'] = Vault.data(
                    self._KEY_OWNER,
                    yt_token_key_name
                )
                ya_make_opts.append('--yt-store')
            if build_with_dist:
                ya_make_opts.append('--dist -E --add-result=".a"')
            cmd = (
                'debuild '
                '--preserve-env '
                '--set-envvar=YA_MAKE_OPTS="{ya_make_opts}" '
                '-k{mail}'
            ).format(
                ya_make_opts=' '.join(ya_make_opts),
                mail=self.Parameters.deb_mail
            )
            sp.check_call(
                cmd,
                shell=True,
                env=environment,
                cwd=module_path,
                stdout=pl.stdout,
                stderr=sp.STDOUT
            )

    def __upload(self, repos, module_path):
        repos_config = {repo: {}
                        for repo in re.sub('[\s+]', '', repos).split(',')}
        with sandbox.projects.common.debpkg.DebRelease(dupload_config=repos_config,
                                                       login=self.Parameters.deb_user) as deb:
            deb.debrelease(
                ['-f', '--to', repos],
                work_dir=module_path
            )

    def __make_unstrip_resource(self, unstrip_dir, package_version, revision):
        market_dir = Path(unstrip_dir).joinpath('market')
        if market_dir.exists():
            root_path_bin = str(market_dir)
            for root, _, file_names in os.walk(root_path_bin):
                for filename in file_names:
                    absolute_bin_path = str(Path(root).joinpath(filename))
                    relative_bin_path = absolute_bin_path.replace(unstrip_dir, '')[1:]
                    is_unstrip_bin = (os.path.isfile(absolute_bin_path)
                                      and os.access(absolute_bin_path, os.X_OK)
                                      and 'test' not in absolute_bin_path)
                    if is_unstrip_bin:
                        res = UnstripBinary(self, filename, absolute_bin_path)
                        res.binary_path = os.path.dirname(relative_bin_path)
                        res.binary_name = os.path.basename(relative_bin_path)
                        res.package_version = package_version
                        res.revision = revision

                        app_res = ResourceData(res)
                        app_res.path.write_bytes(Path(absolute_bin_path).read_bytes())
                        app_res.ready()

    def on_execute(self):
        if self.Parameters.branch_path:
            url = svn.Arcadia.branch_url(
                branch=self.Parameters.branch_path,
                revision=self.Parameters.revision
            )
        else:
            url = svn.Arcadia.trunk_url(
                revision=self.Parameters.revision
            )

        arcadia_root = svn.Arcadia.get_arcadia_src_dir(
            url=url,
            copy_trunk=True
        )
        module_path = str(Path(arcadia_root).joinpath(self.Parameters.module_path))
        unstrip_dir = str(self.path())

        with Key(self, self._KEY_OWNER, self.Parameters.ssh_key_name):
            with GpgKey(self.Parameters.gpg_private_key_name, self.Parameters.gpg_public_key_name):
                self.__make_debian_changelog(
                    version=self.Parameters.package_version,
                    module_path=module_path
                )
                self.__debuild(
                    module_path=module_path,
                    run_test=self.Parameters.run_tests,
                    out_dir=unstrip_dir,
                    yt_token_key_name=self.Parameters.yt_token_key_name,
                    build_with_dist=self.Parameters.build_with_dist
                )

                if self.Parameters.dist_repos:
                    self.__upload(
                        repos=str(self.Parameters.dist_repos),
                        module_path=module_path
                    )

        if self.Parameters.generate_unstrip_binary:
            self.__make_unstrip_resource(
                unstrip_dir=unstrip_dir,
                package_version=self.Parameters.package_version,
                revision=self.Parameters.revision
            )
