# coding: utf8
from __future__ import absolute_import, division, print_function, unicode_literals

from sandbox import sdk2
from sandbox.common.types.misc import DnsType
from sandbox.projects.common import binary_task
from sandbox.projects.rasp.utils.email_notifications import EmailNotificationMixin, use_email_notification_params
from sandbox.projects.rasp.utils.juggler import JugglerNotificationMixin, use_juggler_notification_params
from sandbox.projects.rasp.utils.rasp_scripts import RaspScriptsTask, use_rasp_scripts_params
from sandbox.projects.rasp.utils.solomon import SolomonProgressMixin


class RaspDynamicPlatforms(
    binary_task.LastBinaryTaskRelease,
    sdk2.Task,
    EmailNotificationMixin,
    JugglerNotificationMixin,
    RaspScriptsTask,
    SolomonProgressMixin
):
    class Requirements(sdk2.Task.Requirements):
        cores = 1
        ram = 1 * 1024
        dns = DnsType.DNS64

        class Caches(sdk2.Requirements.Caches):  # enable multislot https://wiki.yandex-team.ru/sandbox/clients/#client-tags-multislot
            pass  # means that task do not use any shared caches

    class Parameters(sdk2.Task.Parameters):
        kill_timeout = 60 * 5

        ext_params = binary_task.binary_release_parameters(stable=True)

        source = sdk2.parameters.String('Script name', required=True, choices=[
            (a,a) for a in ('update_dynamic_platforms', 'clean_dynamic_platforms')
        ])

        _rasp_scripts_params = use_rasp_scripts_params()
        _email_notification_params = use_email_notification_params()
        _juggler_notification_params = use_juggler_notification_params()

    def on_save(self):
        binary_task.LastBinaryTaskRelease.on_save(self)
        self.add_email_notifications()
        self.add_juggler_notifications(
            service='{}.{}'.format(self.type, self.Parameters.source),
            environment=self.Parameters.env_type
        )

    def on_execute(self):
        self.set_common_environ()

        from travel.rasp.tasks.platforms import clean_dynamic_platforms, update_dynamic_platforms

        scripts = {
            'clean_dynamic_platforms': clean_dynamic_platforms.lock_aware_run,
            'update_dynamic_platforms': update_dynamic_platforms.lock_aware_run,
        }

        with self.report_progress(self.Parameters.source, self.Parameters.env_type):
            scripts[self.Parameters.source]()
