# -*- coding: utf-8 -*-
"""
Release machine for Fury RtXaron graph
"""
from sandbox.projects.release_machine.components import configs
from sandbox.projects.release_machine.core import const as rm_const
from sandbox.projects.release_machine.core import releasable_items as ri
import sandbox.projects.release_machine.components.job_graph.job_arrows as jg_arrows
import sandbox.projects.release_machine.components.job_graph.job_data as jg_data
import sandbox.projects.release_machine.components.job_graph.job_triggers as jg_triggers
import sandbox.projects.release_machine.components.job_graph.stages.build_stage as jg_build
import sandbox.projects.release_machine.components.job_graph.stages.release_stage as jg_release
import sandbox.projects.release_machine.components.job_graph.stages.pre_release_stage as jg_prerelease
import sandbox.projects.xurma.rm as xurma_rm


ALL_RESOURCES = {
    "FURY_RT_XARON_CONFIGS": "configs/rtxaron.json",
    "FURY_RT_XARON_CONTROL_NODE": "control-nodes/rtxaron.json",
    "FURY_RT_XARON_DATA_NODE": "data-nodes/rtxaron.json",
}


class FuryRtXaronCfg(configs.ReferenceBranchedConfig):
    name = "fury_rtxaron"
    responsible = xurma_rm.RESPONSIBLE

    class ChangelogCfg(configs.ReferenceBranchedConfig.ChangelogCfg):
        dirs = [
            "arcadia/quality/antifraud/xurma",
        ]
        wiki_page = "JandeksPoisk/Antirobots/Xurma/releases/releases-rtxaron/"
        wiki_page_owner = xurma_rm.WIKI_PAGE_OWNER

    class Releases(configs.ReferenceBranchedConfig.Releases):
        @property
        def releasable_items(self):
            return [
                ri.ReleasableItem(
                    name=resource.lower(),
                    data=ri.SandboxResourceData(resource),
                    deploy_infos=[
                        xurma_rm.ya_deploy_stage("fury-rtxaron-test", stage=rm_const.ReleaseStatus.testing),
                        xurma_rm.ya_deploy_stage("fury-rtxaron-preprod", stage=rm_const.ReleaseStatus.prestable),
                        xurma_rm.ya_deploy_stage("fury-rtxaron-prod", stage=rm_const.ReleaseStatus.stable),
                    ],
                ) for resource in ALL_RESOURCES
            ]

        allow_robots_to_release_stable = True

    class Testenv(configs.ReferenceBranchedConfig.Testenv):
        trunk_task_owner = xurma_rm.TRUNK_TASK_OWNER
        testenv_db_owners = xurma_rm.TESTENV_DB_OWNERS

        class JobGraph(configs.ReferenceBranchedConfig.Testenv.JobGraph):
            @property
            def _prerelease(self):
                """
                    Prerelease part of job graph for Testenv with arrows and dependencies
                    :return: list with jobs
                """
                return [
                    jg_prerelease.JobGraphElementNewBranch(),
                    jg_prerelease.JobGraphElementPreliminaryChangelogBranched(),
                    xurma_rm.JobGraphElementChangelogProcessor(ctx={
                        "graph_name": "rtxaron",
                    }),
                    jg_prerelease.JobGraphElementCloneDb(),
                    jg_prerelease.JobGraphElementStartrekBase(job_arrows=(
                        jg_arrows.JobTrigger(
                            "CHANGELOG_PROCESSOR",
                            jg_data.ParentDataResource("changelog", "RELEASE_MACHINE_CHANGELOG"),
                        ),
                    )),
                    jg_prerelease.JobGraphElementWikiBase(job_arrows=(
                        jg_arrows.JobTrigger(
                            "CHANGELOG_PROCESSOR",
                            jg_data.ParentDataResource("changelog", "RELEASE_MACHINE_CHANGELOG"),
                        ),
                        jg_triggers.JobTriggerStartrek(),
                    )),
                    jg_prerelease.JobGraphElementActionPreReleaseStartrekWiki(),
                ]

            @property
            def _branch_part(self):
                branch_part = super(self.__class__, self)._branch_part
                branch_part.append(
                    jg_build.JobGraphElementBuildBranched(
                        task_name="FURY_BUILD_RTXARON",
                        out={resource: 30 for resource in ALL_RESOURCES},
                        ctx={
                            "resource_type": ";".join(ALL_RESOURCES.keys()),
                            "packages": ";".join(
                                "quality/antifraud/xurma/deploy/fury/" + path
                                for path in ALL_RESOURCES.values()
                            ),
                        },
                    )
                )
                return branch_part

            @property
            def _release(self):
                release_part = super(self.__class__, self)._release
                for release_to in (
                    rm_const.ReleaseStatus.testing,
                    rm_const.ReleaseStatus.prestable,
                    rm_const.ReleaseStatus.stable,
                ):
                    release_part.append(jg_release.JobGraphElementReleaseBranched(
                        release_to=release_to,
                        job_arrows=[
                            jg_triggers.JobTriggerBuild(
                                parent_job_data=[
                                    jg_data.ParentDataDict(
                                        input_key="component_resources",
                                        dict_key=resource.lower(),
                                        resource_name=resource,
                                    ) for resource in ALL_RESOURCES
                                ],
                            ),
                        ],
                    ))
                    release_part.append(jg_release.JobGraphElementActionReleaseBranched(
                        release_to=release_to,
                    ))
                return release_part

    class SvnCfg(configs.ReferenceBranchedConfig.SvnCfg):
        branch_name = "xurma/graphs/rtxaron"
        tag_name = "xurma/graphs/rtxaron"

    class Notify(configs.ReferenceBranchedConfig.Notify):

        class Startrek(configs.ReferenceBranchedConfig.Notify.Startrek):
            queue = xurma_rm.STARTREK_QUEUE
            assignee = xurma_rm.STARTREK_ASSIGNEE
            ticket_type = "Release"
            summary_template = u"Fury RtXaron graph {}"
            add_commiters_as_followers = False
            followers = []
            workflow = xurma_rm.STARTREK_WORKFLOW
