# -*- coding: utf-8 -*-
"""
Release machine for Fury Supermod graph
"""
from sandbox.projects.release_machine.components import configs
from sandbox.projects.release_machine.core import const as rm_const
from sandbox.projects.release_machine.core import releasable_items as ri
import sandbox.projects.release_machine.components.job_graph.job_arrows as jg_arrows
import sandbox.projects.release_machine.components.job_graph.job_data as jg_data
import sandbox.projects.release_machine.components.job_graph.job_triggers as jg_triggers
import sandbox.projects.release_machine.components.job_graph.stages.build_stage as jg_build
import sandbox.projects.release_machine.components.job_graph.stages.release_stage as jg_release
import sandbox.projects.release_machine.components.job_graph.stages.pre_release_stage as jg_prerelease
import sandbox.projects.xurma.rm as xurma_rm


NANNY_RESOURCES = {
    "FURY_SUPER_MOD_DATA": "graphs/supermod.json",
}
YA_DEPLOY_RESOURCES = {
    "FURY_SUPER_MOD_CONFIGS": "configs/supermod.json",
    "FURY_SUPER_MOD_CONTROL_NODE": "control-nodes/supermod.json",
    "FURY_SUPER_MOD_DATA_NODE": "data-nodes/supermod.json",
}
ALL_RESOURCES = xurma_rm.merge_dicts(NANNY_RESOURCES, YA_DEPLOY_RESOURCES)


class FurySupermodCfg(configs.ReferenceBranchedConfig):
    name = "fury_supermod"
    responsible = xurma_rm.RESPONSIBLE

    class ChangelogCfg(configs.ReferenceBranchedConfig.ChangelogCfg):
        dirs = [
            "arcadia/quality/antifraud/xurma",
        ]
        wiki_page = "JandeksPoisk/Antirobots/Xurma/releases/releases-supermod/"
        wiki_page_owner = xurma_rm.WIKI_PAGE_OWNER

    class Releases(configs.ReferenceBranchedConfig.Releases):
        @property
        def releasable_items(self):
            return [
                ri.ReleasableItem(
                    name=resource.lower(),
                    data=ri.SandboxResourceData(resource),
                    deploy_infos=[
                        ri.NannyDeployInfo([
                            ri.DeployService("prod_supermod_updater_fury_yp"),
                        ]),
                    ],
                ) for resource in NANNY_RESOURCES
            ] + [
                ri.ReleasableItem(
                    name=resource.lower(),
                    data=ri.SandboxResourceData(resource),
                    deploy_infos=[
                        xurma_rm.ya_deploy_stage("fury-supermod-preprod", stage=rm_const.ReleaseStatus.prestable),
                    ],
                ) for resource in YA_DEPLOY_RESOURCES
            ]

        allow_robots_to_release_stable = True

    class Testenv(configs.ReferenceBranchedConfig.Testenv):
        trunk_task_owner = xurma_rm.TRUNK_TASK_OWNER
        testenv_db_owners = xurma_rm.TESTENV_DB_OWNERS

        class JobGraph(configs.ReferenceBranchedConfig.Testenv.JobGraph):
            @property
            def _prerelease(self):
                """
                    Prerelease part of job graph for Testenv with arrows and dependencies
                    :return: list with jobs
                """
                return [
                    jg_prerelease.JobGraphElementNewBranch(),
                    jg_prerelease.JobGraphElementPreliminaryChangelogBranched(),
                    xurma_rm.JobGraphElementChangelogProcessor(ctx={
                        "graph_name": "supermod",
                    }),
                    jg_prerelease.JobGraphElementCloneDb(),
                    jg_prerelease.JobGraphElementStartrekBase(job_arrows=(
                        jg_arrows.JobTrigger(
                            "CHANGELOG_PROCESSOR",
                            jg_data.ParentDataResource("changelog", "RELEASE_MACHINE_CHANGELOG"),
                        ),
                    )),
                    jg_prerelease.JobGraphElementWikiBase(job_arrows=(
                        jg_arrows.JobTrigger(
                            "CHANGELOG_PROCESSOR",
                            jg_data.ParentDataResource("changelog", "RELEASE_MACHINE_CHANGELOG"),
                        ),
                        jg_triggers.JobTriggerStartrek(),
                    )),
                    jg_prerelease.JobGraphElementActionPreReleaseStartrekWiki(),
                ]

            @property
            def _branch_part(self):
                branch_part = super(self.__class__, self)._branch_part
                branch_part.append(
                    jg_build.JobGraphElementBuildBranched(
                        task_name="FURY_BUILD_SUPERMOD",
                        out={resource: 30 for resource in ALL_RESOURCES},
                        ctx={
                            "resource_type": ";".join(ALL_RESOURCES.keys()),
                            "packages": ";".join(
                                "quality/antifraud/xurma/deploy/fury/" + path
                                for path in ALL_RESOURCES.values()
                            ),
                        },
                    )
                )
                return branch_part

            @property
            def _release(self):
                release_part = super(self.__class__, self)._release
                for release_to in (
                    rm_const.ReleaseStatus.testing,
                    rm_const.ReleaseStatus.prestable,
                    rm_const.ReleaseStatus.stable,
                ):
                    release_part.append(jg_release.JobGraphElementReleaseBranched(
                        release_to=release_to,
                        job_arrows=[
                            jg_triggers.JobTriggerBuild(
                                parent_job_data=[
                                    jg_data.ParentDataDict(
                                        input_key="component_resources",
                                        dict_key=resource.lower(),
                                        resource_name=resource,
                                    ) for resource in ALL_RESOURCES
                                ],
                            ),
                        ],
                    ))
                # stable and testing
                for release_to in (rm_const.ReleaseStatus.testing, rm_const.ReleaseStatus.stable):
                    release_part.append(jg_release.JobGraphElementActionReleaseBranched(
                        release_to=release_to,
                    ))
                # prestable
                release_part.append(xurma_rm.JobGraphElementAgreementMaker(
                    job_name_parameter="prestable",
                    ctx={
                        "text": u"Согласование на приемку релиза",
                        "approvers": ["granilace,limixis,s-kulivets"]
                    }
                ))
                release_part.append(jg_release.JobGraphElementActionReleaseBranched(
                    release_to=rm_const.ReleaseStatus.prestable,
                    job_arrows=jg_arrows.JobTrigger("AGREEMENT_MAKER"),
                ))
                return release_part

    class SvnCfg(configs.ReferenceBranchedConfig.SvnCfg):
        branch_name = "xurma/graphs/supermod"
        tag_name = "xurma/graphs/supermod"

    class Notify(configs.ReferenceBranchedConfig.Notify):

        class Startrek(configs.ReferenceBranchedConfig.Notify.Startrek):
            queue = xurma_rm.STARTREK_QUEUE
            assignee = xurma_rm.STARTREK_ASSIGNEE
            ticket_type = "Release"
            summary_template = u"Fury Supermod graph {}"
            add_commiters_as_followers = False
            followers = [assignee, "granilace", "limixis", "s-kulivets"]
            workflow = xurma_rm.STARTREK_WORKFLOW
