# -*- coding: utf-8 -*-

import logging

import sandbox.projects.rtmr.common as rtmr_common
import sandbox.common.types.resource as ctr
from sandbox import common
from sandbox import sdk2
from sandbox.projects.rtmr.RtmrBuildCommonUserdata import RtmrBuildCommonUserdata
from sandbox.projects.rtmr.RtmrBuildUsertask import RtmrBuildUsertask
from sandbox.projects.rtmr.RtmrUserDeploy import RtmrUserDeploy
from sandbox.projects.rtmr.RtmrDiffUsertaskConfig import RtmrDiffUsertaskConfig
from sandbox.projects.rtmr.clusters import RTMR_CLUSTERS
from sandbox.projects.rtmr.resources import RtmrConfigTool


class RtmrUserRelease(sdk2.Task):
    """Release RTMR user components"""

    class Requirements(sdk2.Task.Requirements):
        cores = 1
        disk_space = 2 * 1024  # 2Gb

        class Caches(sdk2.Requirements.Caches):
            pass

    class Parameters(sdk2.Task.Parameters):
        description = "Release RTMR user components"
        kill_timeout = 4 * 60 * 60

        with sdk2.parameters.CheckGroup("Clusters") as clusters:
            for _name in RTMR_CLUSTERS:
                clusters.values[_name] = _name

        build_userdata = sdk2.parameters.Bool("Build userdata", default_value=False)
        with build_userdata.value[True]:
            with sdk2.parameters.Group("Build userdata parameters"):
                statface_username = sdk2.parameters.String("Statface username")
                statface_password_vault_owner = sdk2.parameters.String(
                    "Vault secret owner with Statface password"
                )
                statface_password_vault = sdk2.parameters.String("Vault secret name with Statface password")

        build_usertasks = sdk2.parameters.Bool("Build usertasks", default_value=False)
        with build_usertasks.value[True]:
            with sdk2.parameters.Group("Build usertasks parameters"):
                arcadia_url = sdk2.parameters.ArcadiaUrl(
                    "Arcadia URL",
                    required=True,
                    default_value="arcadia:/arc/trunk/arcadia"
                )
                task_ids = sdk2.parameters.String("Task ids to build (task1,task2... empty for all)")
                parallel_limit = sdk2.parameters.Integer(
                    "Parallel build package limit",
                    required=True,
                    default_value=99
                )
                build_sys = sdk2.parameters.Bool("Include system tasks to build", default_value=True)
                strip_packages = sdk2.parameters.Bool("Strip debug info", default_value=True)
                create_debug_packages = sdk2.parameters.Bool("Create debug info packages", default_value=False)
                account_names = sdk2.parameters.String("Names of accounts for tasks that need to be built")
                publish_packages = sdk2.parameters.Bool("Publish packages", default=True)

        with build_userdata.value[False]:
            add_last_userdata_resource = sdk2.parameters.Bool(
                "Add last common userdata resource if not building userdata",
                default_value=False)

        with sdk2.parameters.Group("Build task settings") as build_settings:
            ya_package_name_for_sign = sdk2.parameters.String("Name of the user to sign the package", required=False)
            build_disk_space = sdk2.parameters.Integer("Execution space (Gb)", default_value=60)

        with sdk2.parameters.Group("Config diff settings") as diff_settings:
            rtmr_config_resource = rtmr_common.LastResource(
                "RTMR Config Tool",
                resource_type=RtmrConfigTool,
                required=False
            )
            wait_successful_diff = sdk2.parameters.Bool(
                "Wait for successful diff",
                default_value=False,
            )

        with sdk2.parameters.Group("YF deploy settings") as yf_deploy_settings:
            deploy_to_yf = sdk2.parameters.Bool(
                "Deploy build to YF",
                required=False,
                default_value=True)

            with deploy_to_yf.value[True]:
                cleanup_first = sdk2.parameters.Bool(
                    "Run rtmr-deploy cleanup before deploying the new account",
                    default=True)

                switch = sdk2.parameters.Bool(
                    "Switch execution to new version",
                    default=True)

                slots = sdk2.parameters.Integer(
                    "Number of slots for YF function",
                    default=None)

                cleanup_missing = sdk2.parameters.Bool(
                    "Delete tasks not present in the generated config",
                    default=False)

                update_previous = sdk2.parameters.Bool(
                    "Update artifacts of previous version instead of full release",
                    default=False)

                oauth_token_name = sdk2.parameters.String(
                    "Vault secret name with rtmr-deploy OAuth token",
                    default="rtmr_deploy_usertask_yf_oauth_token",
                    required=True)

                notification_logins = sdk2.parameters.String(
                    "Comma separated list of staff logins for crash notifications (by e-mail)",
                    required=False,
                    default=None)

        task_config_exclude_accounts_per_cluster = sdk2.parameters.Dict(
            "Accounts to exclude from task config building per cluster",
            required=False,
            default={_name: "" for _name in RTMR_CLUSTERS})

    class Context(sdk2.Task.Context):
        arcadia_branch = None
        arcadia_revision = None
        build_usertasks_taskid = None
        build_userdata_taskid = None
        rtmr_config = None
        diff_taskids = list()
        deploy_yf_taskids_per_cluster = dict()
        user_deploy_taskids_per_cluster = dict()

    def get_version_from_url(self, arcadia_url):
        parsed_url = sdk2.svn.Arcadia.parse_url(arcadia_url)
        if parsed_url.branch is None:
            version = "trunk"
        else:
            version = parsed_url.branch
        if parsed_url.revision is not None:
            version += "@" + parsed_url.revision
        return version

    def build_userdata(self):
        self.set_info("Build common userdata")
        build_task = RtmrBuildCommonUserdata(
            self,
            description="Build RTMR userdata",
            build_disk_space=self.Parameters.build_disk_space,
            statface_username=self.Parameters.statface_username,
            statface_password_vault_owner=self.Parameters.statface_password_vault_owner,
            statface_password_vault=self.Parameters.statface_password_vault,
            ya_package_name_for_sign=self.Parameters.ya_package_name_for_sign,
        )
        build_task.save().enqueue()
        self.Context.build_userdata_taskid = build_task.id

    def build_usertasks(self):
        self.set_info("Build usertasks and config")
        params = dict()
        task_ids = []
        if self.Parameters.account_names:
            params["account_names"] = self.Parameters.account_names
        else:
            if self.Parameters.task_ids:
                task_ids.append(self.Parameters.task_ids)
            if len(task_ids) > 0:
                params["task_ids"] = ",".join(task_ids)

        params["build_sys"] = self.Parameters.build_sys

        build_task = RtmrBuildUsertask(
            self,
            description="Build RTMR usertasks " + self.get_version_from_url(self.Parameters.arcadia_url),
            priority=self.Parameters.priority,
            parallel_limit=self.Parameters.parallel_limit,
            clusters=self.get_clusters(),
            strip_packages=self.Parameters.strip_packages,
            create_debug_packages=self.Parameters.create_debug_packages,
            build_disk_space=self.Parameters.build_disk_space,
            ya_package_name_for_sign=self.Parameters.ya_package_name_for_sign,
            arcadia_url=sdk2.svn.Arcadia.replace(
                self.Parameters.arcadia_url,
                revision=self.Context.arcadia_revision
            ),
            publish_packages=self.Parameters.publish_packages,
            task_config_exclude_accounts_per_cluster=self.Parameters.task_config_exclude_accounts_per_cluster,
            **params
        )
        build_task.save().enqueue()
        self.Context.build_usertasks_taskid = build_task.id

    def get_build_tasks(self):
        tasks = list()
        tasks.append(self.Context.build_userdata_taskid)
        tasks.append(self.Context.build_usertasks_taskid)
        logging.info("Build tasks %r", tasks)
        return filter(lambda task: task is not None, tasks)

    def get_clusters(self):
        clusters = set(self.Parameters.clusters)
        return list(clusters)

    def get_deploy_yf_tasks(self):
        tasks = list()
        tasks += list(self.Context.deploy_yf_taskids_per_cluster.values())
        logging.info("Deploy to YF tasks: %r", tasks)
        return filter(lambda task: task is not None, tasks)

    def start_diff(self):
        tasks = dict()
        for cluster_name in self.get_clusters():
            task = RtmrDiffUsertaskConfig(
                self,
                description="Config diff for cluster " + cluster_name,
                priority=self.Parameters.priority,
                cluster_name=cluster_name,
                config_task=sdk2.Task[self.Context.build_usertasks_taskid],
                rtmr_config_resource=sdk2.Resource[self.Context.rtmr_config],
            )
            task.save().enqueue()
            self.Context.diff_taskids.append(task.id)
            self.Context.save()
            tasks[cluster_name] = task.id
        messages = [
            "{cluster} <a href=\"https://sandbox.yandex-team.ru/task/{taskid}/view\">{taskid}</a>".format(
                cluster=cluster_name,
                taskid=taskid
            )
            for (cluster_name, taskid) in tasks.iteritems()
        ]
        if len(messages) > 0:
            self.set_info("Diff tasks:\n " + "\n ".join(messages), do_escape=False)

    def make_userdeploy_tasks(self):
        messages = []
        for cluster_name in self.get_clusters():
            params = dict()
            if not self.Parameters.build_userdata:
                params["add_last_userdata_resource"] = self.Parameters.add_last_userdata_resource

            accounts = self.Parameters.account_names.split(",")
            accounts = filter(bool, [account.strip() for account in accounts])

            for account in accounts:
                task = RtmrUserDeploy(
                    self,
                    description="Deploy usertasks to " + cluster_name,
                    priority=self.Parameters.priority,
                    cluster=cluster_name,
                    account=account,
                    build_task=self,
                    deploy_to_yf=self.Parameters.deploy_to_yf,
                    rtmr_config_resource=self.Parameters.rtmr_config_resource,
                    cleanup_first=self.Parameters.cleanup_first,
                    switch=self.Parameters.switch,
                    slots=self.Parameters.slots,
                    cleanup_missing=self.Parameters.cleanup_missing,
                    update_previous=self.Parameters.update_previous,
                    oauth_token_name=self.Parameters.oauth_token_name,
                    notification_logins=self.Parameters.notification_logins,
                    **params
                )
                task.save()
                self.Context.user_deploy_taskids_per_cluster[cluster_name] = task.id
                self.Context.save()

                messages.append(
                    "{cluster}, {account}: <a href=\"https://sandbox.yandex-team.ru/task/{taskid}/view\">{taskid}</a>".format(
                        cluster=cluster_name, account=account, taskid=task.id))

        if len(messages) > 0:
            self.set_info("User deploy tasks:\n " + "\n".join(messages), do_escape=False)

    def on_execute(self):
        if not (self.Parameters.build_userdata or
                self.Parameters.build_usertasks):
            raise common.errors.TaskError("Need to specify build tasks")

        with self.memoize_stage.update_config_tool_resource(commit_on_entrance=False):
            if self.Parameters.rtmr_config_resource is not None:
                self.Context.rtmr_config = self.Parameters.rtmr_config_resource.id
            else:
                self.Context.rtmr_config = sdk2.Resource.find(resource_type=RtmrConfigTool, state=ctr.State.READY)\
                    .order(-sdk2.Resource.id)\
                    .first().id

        # Determine proper revision
        with self.memoize_stage.determine_revision(commit_on_entrance=False):
            rtmr_common.update_version_info(self, self.Parameters.arcadia_url)

        # Start build tasks
        with self.memoize_stage.build_userdata(commit_on_entrance=False):
            if self.Parameters.build_userdata:
                self.build_userdata()

        with self.memoize_stage.build_usertasks(commit_on_entrance=False):
            if self.Parameters.build_usertasks:
                self.build_usertasks()

        # Wait build tasks
        with self.memoize_stage.wait_builds(commit_on_entrance=False, commit_on_wait=False):
            rtmr_common.wait_tasks(self.get_build_tasks())

        # Create user deploy tasks
        if self.Parameters.deploy_to_yf:
            with self.memoize_stage.user_deploy(commit_on_entrance=False):
                self.make_userdeploy_tasks()

        # Start diff tasks
        if self.Parameters.build_usertasks:
            with self.memoize_stage.show_config_diff(commit_on_entrance=False):
                self.start_diff()

            if self.Parameters.wait_successful_diff:
                with self.memoize_stage.wait_succesful_diff(commit_on_entrance=False, commit_on_wait=False):
                    rtmr_common.wait_tasks(self.Context.diff_taskids, stop_when_fail=False)
